==> secretflow-1.10.0b0-cp311-cp311-manylinux_2_28_aarch64.whl: secretflow/distributed/primitive.py <==
# Copyright 2023 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import inspect
import logging
import multiprocessing
from enum import Enum, unique
from functools import partial
from typing import List, Union

import fed as rayfed
import jax
import ray
from ray import Language
from ray._private import ray_option_utils
from ray.actor import ActorClass, _inject_tracing_into_class, ray_constants
from ray.remote_function import RemoteFunction

import secretflow.ic.remote as ic
from secretflow.distributed import fed as sf_fed
from secretflow.distributed.memory_api import mem_remote
from secretflow.utils.ray_compatibility import ray_version_less_than_2_0_0


@unique
class DISTRIBUTION_MODE(Enum):
    PRODUCTION = 1
    SIMULATION = 2
    DEBUG = 3
    INTERCONNECTION = 4
    RAY_PRODUCTION = 5


_distribution_mode = DISTRIBUTION_MODE.SIMULATION
_is_cluster_active = False
_current_cluster_id = -1


def set_distribution_mode(mode: DISTRIBUTION_MODE):
    global _distribution_mode
    if mode in DISTRIBUTION_MODE:
        _distribution_mode = mode
    else:
        raise Exception(f"Illegal distribute mode, only support ({DISTRIBUTION_MODE})")


def get_distribution_mode():
    global _distribution_mode
    return _distribution_mode


def get_current_cluster_idx():
    """
    Get the current secretflow cluster index.
    In general situation, users will execute sf.init() at the first of their codes,
    and execute sf.shutdown() at the end.
    But in some cases, the following code may appear:
        - sf.init()
        - do_something_in_sf()
        - sf.shutdown()
        - do_something_else()
        - sf.init()
        - do_something_in_sf()
        - sf.shutdown()
    To ensure some variable can work as expect when it across different clusters,
    we use an index to indicate diferent active cluster.
    As shown above, between the first sf.init() and sf.shutdown(), the cluster index is 0,
    and the between the second sf.init() and sf.shutdown(), the cluster index is 1,
    In do_something_else(), the cluster index is -1 (no active cluster).

    Returns: The current cluster id, -1 for not active.

    """
    global _current_cluster_id
    global _is_cluster_active
    if _is_cluster_active:
        return _current_cluster_id
    else:
        return -1


def active_sf_cluster():
    """
    Record the cluster's active status and generate the current cluster index.
    """
    global _current_cluster_id
    global _is_cluster_active
    _current_cluster_id = _current_cluster_id + 1
    _is_cluster_active = True


def get_cluster_avaliable_resources():
    if get_distribution_mode() == DISTRIBUTION_MODE.DEBUG:
        from secretflow.device import global_state

        cpu_counts = multiprocessing.cpu_count()
        avaliable_resources = {party: cpu_counts for party in global_state.parties()}
        avaliable_resources['CPU'] = cpu_counts
        return avaliable_resources
    elif get_distribution_mode() == DISTRIBUTION_MODE.SIMULATION:
        return ray.available_resources()
    else:
        raise NotImplementedError(
            "Need but not implement. This will be used by tune when production mode."
        )


def _is_cython(obj):
    """Check if an object is a Cython function or method"""

    # TODO(suo): We could split these into two functions, one for Cython
    # functions and another for Cython methods.
    # TODO(suo): There doesn't appear to be a Cython function 'type' we can
    # check against via isinstance. Please correct me if I'm wrong.
    def check_cython(x):
        return type(x).__name__ == "cython_function_or_method"

    # Check if function or method, respectively
    return check_cython(obj) or (
        hasattr(obj, "__func__") and check_cython(obj.__func__)
    )


def remote(*args, **kwargs):
    if get_distribution_mode() == DISTRIBUTION_MODE.PRODUCTION:
        return sf_fed.remote(*args, **kwargs)
    elif get_distribution_mode() == DISTRIBUTION_MODE.RAY_PRODUCTION:
        return rayfed.remote(*args, **kwargs)
    elif get_distribution_mode() == DISTRIBUTION_MODE.SIMULATION:
        return ray_remote(*args, **kwargs)
    elif get_distribution_mode() == DISTRIBUTION_MODE.DEBUG:
        return mem_remote(*args, **kwargs)
    elif get_distribution_mode() == DISTRIBUTION_MODE.INTERCONNECTION:
        return ic.remote(*args, **kwargs)
    else:
        raise Exception(f"Illegal distribute mode, only support ({DISTRIBUTION_MODE})")


def get(
    object_refs: Union[
        Union[ray.ObjectRef, List[ray.ObjectRef]],
        Union[sf_fed.FedObject, List[sf_fed.FedObject]],
        Union[rayfed.FedObject, List[rayfed.FedObject]],
        Union[object, List[object]],
    ]
):
    if get_distribution_mode() == DISTRIBUTION_MODE.PRODUCTION:
        return sf_fed.get(object_refs)
    elif get_distribution_mode() == DISTRIBUTION_MODE.RAY_PRODUCTION:
        return rayfed.get(object_refs)
    elif get_distribution_mode() == DISTRIBUTION_MODE.SIMULATION:
        return ray.get(object_refs)
    elif get_distribution_mode() == DISTRIBUTION_MODE.DEBUG:
        return object_refs
    elif get_distribution_mode() == DISTRIBUTION_MODE.INTERCONNECTION:
        return ic.get(object_refs)
    else:
        raise Exception(f"Illegal distribute mode, only support ({DISTRIBUTION_MODE})")


def kill(actor, *, no_restart=True):
    if get_distribution_mode() == DISTRIBUTION_MODE.PRODUCTION:
        pass
    elif get_distribution_mode() == DISTRIBUTION_MODE.RAY_PRODUCTION:
        return rayfed.kill(actor, no_restart=no_restart)
    elif get_distribution_mode() == DISTRIBUTION_MODE.SIMULATION:
        return ray.kill(actor, no_restart=no_restart)
    elif get_distribution_mode() == DISTRIBUTION_MODE.DEBUG:
        logging.warning("Actor be killed")
    elif get_distribution_mode() == DISTRIBUTION_MODE.INTERCONNECTION:
        logging.warning("Actor be killed")
    else:
        raise Exception(f"Illegal distribute mode, only support ({DISTRIBUTION_MODE})")


def shutdown(on_error=False):
    """Shutdown the secretflow environment.

    Args:
        on_error: optional; this is useful only in production mode (using RayFed).
            This parameter indicates whether an error has occurred on your main
            thread. Rayfed is desigend to reliably send all data to peers, but will
            cease transmission if an error is detected. However, Rayfed is not equipped
            to automatically identify errors under all circumstances, particularly
            those that affect only one party independently of others. Should you
            encounter such an error, please notify Rayfed upon shutdown, and it will
            discontinue any ongoing data transmissions if
            `continue_waiting_for_data_sending_on_error` is not True.
    """
    global _is_cluster_active
    _is_cluster_active = False
    if get_distribution_mode() == DISTRIBUTION_MODE.PRODUCTION:
        sf_fed.shutdown(on_error=on_error)
    elif get_distribution_mode() == DISTRIBUTION_MODE.RAY_PRODUCTION:
        rayfed.shutdown(on_error=on_error)
        ray.shutdown(_exiting_interpreter=True)
    elif get_distribution_mode() == DISTRIBUTION_MODE.SIMULATION:
        ray.shutdown()
    elif get_distribution_mode() == DISTRIBUTION_MODE.DEBUG:
        logging.warning("Shut Down!")
    elif get_distribution_mode() == DISTRIBUTION_MODE.INTERCONNECTION:
        logging.warning("Shut Down!")
    else:
        raise Exception(f"Illegal distribute mode, only support ({DISTRIBUTION_MODE})")


def _resolve_args(*args, **kwargs):
    arg_flat, arg_tree = jax.tree_util.tree_flatten((args, kwargs))
    refs = {
        pos: arg for pos, arg in enumerate(arg_flat) if isinstance(arg, ray.ObjectRef)
    }

    actual_vals = ray.get(list(refs.values()))
    for pos, actual_val in zip(refs.keys(), actual_vals):
        arg_flat[pos] = actual_val

    args, kwargs = jax.tree_util.tree_unflatten(arg_tree, arg_flat)
    return args, kwargs


class RemoteFunctionWrapper(RemoteFunction):
    def _remote(self, *args, **kwargs):
        args, kwargs = _resolve_args(*args, **kwargs)
        return super()._remote(*args, **kwargs)

    def party(self, party: str):
        self.party = party
        if 'resources' in self._default_options:
            self._default_options['resources'].update({self.party: 1})
        else:
            self._default_options.update({'resources': {self.party: 1}})
        return self

    def options(self, **task_options):
        if hasattr(self, 'party') and self.party:
            if 'resources' in task_options:
                task_options['resources'].update({self.party: 1})
            else:
                task_options.update({'resources': {self.party: 1}})
        return super().options(**task_options)


class ActorClassWrapper(ActorClass):
    def party(self, party: str):
        self.party = party
        if 'resources' in self._default_options:
            self._default_options['resources'].update({self.party: 1})
        else:
            self._default_options.update({'resources': {self.party: 1}})
        return self

    def options(self, **actor_options):
        if hasattr(self, 'party') and self.party:
            if 'resources' in actor_options:
                actor_options['resources'].update({self.party: 1})
            else:
                actor_options.update({'resources': {self.party: 1}})
        return super().options(**actor_options)

    def remote(self, *args, **kwargs):
        args, kwargs = _resolve_args(*args, **kwargs)
        return super().remote(*args, **kwargs)


def _make_actor(cls, actor_options):
    if ray_version_less_than_2_0_0():
        from ray import ActorClassID
        from ray.actor import modify_class as _modify_class
    else:
        from ray.actor import ActorClassID, _modify_class

    Class = _modify_class(cls)
    _inject_tracing_into_class(Class)

    if "max_restarts" in actor_options:
        if actor_options["max_restarts"] != -1:  # -1 represents infinite restart
            # Make sure we don't pass too big of an int to C++, causing
            # an overflow.
            actor_options["max_restarts"] = min(
                actor_options["max_restarts"], ray_constants.MAX_INT64_VALUE
            )

    return ActorClassWrapper._ray_from_modified_class(
        Class,
        ActorClassID.from_random(),
        actor_options,
    )


def _make_remote(function_or_class, options):
    if inspect.isfunction(function_or_class) or _is_cython(function_or_class):
        ray_option_utils.validate_task_options(options, in_options=False)
        return RemoteFunctionWrapper(
            Language.PYTHON,
            function_or_class,
            None,
            options,
        )
    if inspect.isclass(function_or_class):
        ray_option_utils.validate_actor_options(options, in_options=False)
        return _make_actor(function_or_class, options)

    raise TypeError(
        "The @ray.remote decorator must be applied to either a function or a class."
    )


def ray_remote(*args, **kwargs):
    if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
        # This is the case where the decorator is just @ray.remote.
        # "args[0]" is the class or function under the decorator.
        return _make_remote(args[0], {})
    assert len(args) == 0 and len(kwargs) > 0, ray_option_utils.remote_args_error_string
    return partial(_make_remote, options=kwargs)


# Whether running in interconnection mode
def in_ic_mode() -> bool:
    return get_distribution_mode() == DISTRIBUTION_MODE.INTERCONNECTION

==> secretflow-1.11.0b1-cp311-cp311-manylinux_2_28_aarch64.whl: secretflow/distributed/primitive.py <==
# Copyright 2023 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Union

from .const import DISTRIBUTION_MODE, FED_OBJECT_TYPES
from .op_context import SFOpContext


_sf_op_context = SFOpContext(DISTRIBUTION_MODE.PRODUCTION)


def set_distribution_mode(mode: DISTRIBUTION_MODE):
    _sf_op_context.set_distribution_mode(mode)


def get_distribution_mode():
    return _sf_op_context.get_distribution_mode()


def get_current_cluster_idx():
    return _sf_op_context.get_current_cluster_idx()


def active_sf_cluster():
    _sf_op_context.active_sf_cluster()


def get_cluster_available_resources():
    return _sf_op_context.get_cluster_available_resources()


def init(mode: DISTRIBUTION_MODE, **kwargs):
    _sf_op_context.set_distribution_mode(mode)
    return _sf_op_context.init(**kwargs)


def remote(*args, **kwargs):
    return _sf_op_context.remote(*args, **kwargs)


def get(
    object_refs: Union[
        FED_OBJECT_TYPES,
        List[FED_OBJECT_TYPES],
        Union[object, List[object]],
    ]
):
    return _sf_op_context.get(object_refs)


def kill(actor, *, no_restart=True):
    return _sf_op_context.kill(actor, no_restart=no_restart)


def shutdown(on_error=False):
    """Shutdown the secretflow environment.

    Args:
        on_error: optional; this is useful only in production mode (using RayFed).
            This parameter indicates whether an error has occurred on your main
            thread. Rayfed is desigend to reliably send all data to peers, but will
            cease transmission if an error is detected. However, Rayfed is not equipped
            to automatically identify errors under all circumstances, particularly
            those that affect only one party independently of others. Should you
            encounter such an error, please notify Rayfed upon shutdown, and it will
            discontinue any ongoing data transmissions if
            `continue_waiting_for_data_sending_on_error` is not True.
    """
    _sf_op_context.deactivate_sf_cluster()
    _sf_op_context.shutdown(on_error=on_error)


# Whether running in interconnection mode
def in_ic_mode() -> bool:
    return get_distribution_mode() == DISTRIBUTION_MODE.INTERCONNECTION
